mod config;
mod modem;
#[allow(dead_code)]
mod command;

use anyhow;
use std::time::Duration;
use std::thread;
use esp_idf_hal::prelude::*;
use esp_idf_hal::peripherals::Peripherals;
use esp_idf_hal::serial;

use mqtt::control::ConnectReturnCode;
use mqtt::packet::{ConnackPacket, ConnectPacket, PublishPacket, QoSWithPacketIdentifier};
use mqtt::{Decodable, Encodable, TopicName};


fn main() -> anyhow::Result<()> {
    esp_idf_sys::link_patches();

    let dp = Peripherals::take().expect("error taking peripherals");

    // LilyGo TTGO T-Call sim800l board serial pins.
    let serial_rx = dp.pins.gpio26;
    let serial_tx = dp.pins.gpio27;

    let serial_pins = serial::Pins {
        tx: serial_tx,
        rx: serial_rx,
        cts: None,
        rts: None,
    };

    // Create the serial and panic with a message ... if we can't create the serial port, then we
    // can't communicate with the sim800l module, hence we don't run anymore.
    let serial: serial::Serial<serial::UART1, _, _> = serial::Serial::new(
        dp.uart1,
        serial_pins,
        serial::config::Config::default().baudrate(Hertz(115200)),
    )?;

    let (tx, rx) = serial.split();
    let mut mdm = modem::Modem::new(tx, rx);

    let modem_pwrkey = dp.pins.gpio4.into_output()?;
    let modem_rst = dp.pins.gpio5.into_output()?;
    let modem_power = dp.pins.gpio23.into_output()?;

    mdm.init(modem_pwrkey, modem_rst, modem_power)?;

    if !mdm.is_gprs_attached()? {
        let _ = mdm.gprs_attach_ap(
            config::A1_GPRS_AP.apn,
            config::A1_GPRS_AP.username,
            config::A1_GPRS_AP.password,
        )?;
    }

    thread::sleep(Duration::from_millis(2000));

    //println!("setting up client TLS cert");
    //let client_cert = include_bytes!("../certs/full-bin.p12");
    //let client_cert_path = "C:\\USER\\fullchain.pem";

    //let _ = mdm.upload_cert(client_cert_path, client_cert)?;
    //let _ = mdm.ssl_set_client_cert(client_cert_path, "t")?;
    //let _ = mdm.fs_list("C:\\USER\\")?;

    loop {
        if mdm.is_gprs_attached()? {
            let _ = mdm.gprs_connect()?;
            for _ in 0..3 {
                let ip_addr = mdm.gprs_status()?;
                if ip_addr.contains("0.0.0.0") {
                    thread::sleep(Duration::from_millis(2000));
                } else {
                    break
                }
            }

            //println!("connecting to server!");
            //if !mdm.tcp_is_ssl_enabled()? {
            //    let _ = mdm.tcp_ssl_enable()?;
            //}
            if mdm.tcp_is_ssl_enabled()? {
                let _ = mdm.tcp_ssl_disable()?;
            }
            
            let client_id = "c36a72df-5bd6-4f9b-995d-059433bc3267";
            let message = "{\"lat\": 20.475370, \"long\": 44.747224}";

            let _ = mdm.tcp_set_quick_mode(false);
            let _ = mdm.tcp_set_manual_receive(true);

            let _ = mdm.tcp_connect("51.158.66.64", 1883)?;

            let mut buf = Vec::new();

            let mut conn = ConnectPacket::new(client_id);
            conn.set_clean_session(true);
            conn.set_keep_alive(0);
            let _ = conn.encode(&mut buf)?;
            let _ = mdm.tcp_send(&mut buf)?;
            buf.clear();

            thread::sleep(Duration::from_millis(1000));

            let size = mdm.tcp_receive_reply_len()?;
            let mut reply = vec![0 as u8; size];
            let _ = mdm.tcp_receive(&mut reply);
            println!("TCP Received: {}", reply.iter().map(|b| char::from(*b)).collect::<String>());

            println!("sending publish packet");
            let packet = PublishPacket::new(
                TopicName::new("bajsevi/test").unwrap(),
                QoSWithPacketIdentifier::Level0,
                message.as_bytes(),
            );
            let _ = packet.encode(&mut buf)?;
            let _ = mdm.tcp_send(&mut buf)?;
            buf.clear();

            let size = mdm.tcp_receive_reply_len()?;
            let mut reply = vec![0 as u8; size];
            let _ = mdm.tcp_receive(&mut reply);
            println!("TCP Received: {}", reply.iter().map(|b| char::from(*b)).collect::<String>());

            let _ = mdm.tcp_close_connection()?;

            break
        }
    }

    Ok(())
}
